/**
 * \file core/spectrumint.cpp
 * \date 17/10/2018
 * \author Olivier Langella
 * \brief integer spectrum (without intensity)
 */


/*
 * DeepProt, Open Modification Search engine for proteomics
 * Copyright (C) 2018  Olivier Langella <olivier.langella@u-psud.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <pappsomspp/types.h>
#include "spectrumint.h"
#include "../utils/utils.h"


SpectrumInt::SpectrumInt(std::size_t id,
                         std::size_t scan_num,
                         std::vector<std::size_t> &integer_spectrum,
                         const QString &title,
                         unsigned int accuracy,
                         double precursor_mz,
                         unsigned int precursor_charge)
{
  m_id                    = id;
  m_scanOrProtoPeptideNum = scan_num;
  m_title                 = title;
  m_precursorCharge       = precursor_charge;

  // total mass calculation from the precursor mass and the charge
  m_precursorMass =
    (precursor_mz * precursor_charge) - (precursor_charge * pappso::MHPLUS);

  // Computation of the required space for masses storage and allocation of the
  // structure
  std::size_t picCount = (2 * accuracy + 1) * integer_spectrum.size();

  m_itemList.reserve(picCount);
  // duplication of the masses according to the accuracy and structure filling
  for(std::size_t mass : integer_spectrum)
    {
      m_itemList.push_back(mass);
      for(unsigned int cpt1 = 0; cpt1 < accuracy; ++cpt1)
        {
          m_itemList.push_back(mass + cpt1 + 1);
          m_itemList.push_back(mass - cpt1 - 1);
        }
    }
  // conversion into a primitive integer array and sorting
  std::sort(m_itemList.begin(), m_itemList.end());

  // keep unique masses
  // TODO ensure unique without this trick
  m_itemList.erase(std::unique(m_itemList.begin(), m_itemList.end()),
                   m_itemList.end());
}


SpectrumInt::SpectrumInt(std::size_t id,
                         std::size_t protopeptide_id,
                         std::vector<double> &peptide_fragments,
                         const QString &title,
                         unsigned int accuracy_factor,
                         double precursor_mz,
                         unsigned int precursor_charge)
{
  m_id                    = id;
  m_scanOrProtoPeptideNum = protopeptide_id;
  m_title                 = title;
  m_precursorCharge       = precursor_charge;

  // total mass calculation from the precursor mass and the charge
  m_precursorMass = precursor_mz;

  // Computation of the required space for masses storage and allocation of the
  // structure
  // std::size_t picCount = (2 * accuracy_value + 1) * peptide_fragments.size();

  m_itemList.reserve(peptide_fragments.size());

  // specint oldPeak = 0;
  for(pappso::pappso_double mz : peptide_fragments)
    {
      std::size_t newPeak =
        Utils::doubleMassRoundToIntegerMass(mz, accuracy_factor);
      m_itemList.push_back(newPeak);
      // oldPeak = newPeak;
    }
  // conversion into a primitive integer array and sorting
  std::sort(m_itemList.begin(), m_itemList.end());

  // keep unique masses
  m_itemList.erase(std::unique(m_itemList.begin(), m_itemList.end()),
                   m_itemList.end());
}

SpectrumInt::SpectrumInt(const SpectrumInt &other) : ItemCart(other)
{

  // m_id                    = other.m_id;
  m_scanOrProtoPeptideNum = other.m_scanOrProtoPeptideNum;
  m_title                 = other.m_title;
  m_precursorCharge       = other.m_precursorCharge;
  m_precursorMass         = other.m_precursorMass;
  // m_massList              = other.m_massList;
}


SpectrumInt &
SpectrumInt::operator=(const SpectrumInt &other)
{

  ItemCart::operator=(other);
  // m_id                    = other.m_id;
  m_scanOrProtoPeptideNum = other.m_scanOrProtoPeptideNum;
  m_title                 = other.m_title;
  m_precursorCharge       = other.m_precursorCharge;
  m_precursorMass         = other.m_precursorMass;
  // m_massList              = other.m_massList;
  return *this;
}

std::vector<std::size_t>::const_iterator
SpectrumInt::begin() const
{
  return m_itemList.begin();
}
std::vector<std::size_t>::const_iterator
SpectrumInt::end() const
{
  return m_itemList.end();
}

const QString &
SpectrumInt::getTitle() const
{
  return m_title;
}

double
SpectrumInt::getMass() const
{
  return m_precursorMass;
}


unsigned int
SpectrumInt::getCharge() const
{
  return m_precursorCharge;
}

std::size_t
SpectrumInt::getScanOrProtoPeptideNum() const
{
  return m_scanOrProtoPeptideNum;
}
const QString
SpectrumInt::toString() const
{
  QString return_str;
  QTextStream out(&return_str);
  out << m_title << Qt::endl;
  for(std::size_t mass : m_itemList)
    {
      out << mass << " ";
    }
  out << Qt::endl;
  return return_str;
}

const std::vector<std::size_t> &
SpectrumInt::getMassList() const
{
  return m_itemList;
}
